from typing import Any, Dict

DEFAULT_SCRUBBED_FIELDS = (
    'password',
    'secret',
    'passwd',
    'api_key',
    'apikey',
    'bk_token',
    'bk_ticket',
    'access_token',
    'auth',
    'credentials',
)


def scrub_data(data: Dict[str, Any]) -> Dict[str, Any]:
    """Scrub the data, mask all sensitive data fields.

    :return: A new dict, with sensitive data masked as "******".
    """
    if not isinstance(data, dict):
        return data

    def _key_is_sensitive(key: str) -> bool:
        """Check if given key is sensitive."""
        for field in DEFAULT_SCRUBBED_FIELDS:
            if field in key.lower():
                return True
        return False

    result: Dict[str, Any] = {}

    # Use a stack to avoid recursion
    stack = [(data, result)]
    while stack:
        current_data, current_result = stack.pop()

        for key, value in current_data.items():
            if _key_is_sensitive(key):
                current_result[key] = '******'
                continue

            # Process nested data by push it to the stack
            if isinstance(value, dict):
                new_dict: Dict[str, Any] = {}
                current_result[key] = new_dict
                stack.append((value, new_dict))
            else:
                current_result[key] = value
    return result



# -*- coding: utf-8 -*-
import json
import logging
import pickle
import time
from typing import Dict

from django.conf import settings
from django.contrib import auth
from django.http import HttpRequest, HttpResponse
from django.utils.deprecation import MiddlewareMixin
from django.utils.encoding import force_str

from bkpaas_auth.backends import UniversalAuthBackend
from bkpaas_auth.core.constants import ACCESS_PERMISSION_DENIED_CODE
from bkpaas_auth.core.exceptions import AccessPermissionDenied

logger = logging.getLogger(__name__)


class CookieLoginMiddleware(MiddlewareMixin):
    """Call auth.login when user credential cookies changes"""

    def process_request(self, request):
        assert hasattr(request, 'session'), (
            "The CookieLoginMiddleware requires session middleware "
            "to be installed. Edit your MIDDLEWARE%s setting to insert "
            "'django.contrib.sessions.middleware.SessionMiddleware' before "
            "'bkpaas_auth.middlewares.CookieLoginMiddleware'."
        ) % ("_CLASSES" if settings.MIDDLEWARE is None else "")

        backend = UniversalAuthBackend()
        credentials = backend.get_credentials(request)

        # No credentials, call logout
        if not credentials:
            auth.logout(request)
            return self.get_response(request)

        if self.should_authenticate(request, backend, credentials):
            try:
                self.authenticate_and_login(request, credentials)
            except AccessPermissionDenied as e:
                resp = HttpResponse(
                    json.dumps({'code': ACCESS_PERMISSION_DENIED_CODE, 'detail': str(e)}),
                    content_type="application/json",
                )
                resp.status_code = 403
                return resp

        return self.get_response(request)

    def should_authenticate(
        self, request: HttpRequest, backend: UniversalAuthBackend, credentials: Dict[str, str]
    ) -> bool:
        """Decide whether to re-authenticate current credentials or not"""
        # Force re-login if credentials is different from last time
        credentials_been_modified = credentials != request.session.get('auth_credentials', {})
        if credentials_been_modified:
            return True

        # Force re-login if token is empty or obsolete
        token = backend.get_token_from_session(request)
        return token is None

    def authenticate_and_login(self, request: HttpRequest, credentials: Dict[str, str]):
        """Authenticate given credentials and do login(or logout if credentials is invalid)

        :params request: Current request object
        :params credentials: user credentials, such as uin/skey pair
        """
        logger.debug('Authenticating credentials...')
        user = auth.authenticate(request=request, auth_credentials=credentials)
        if user is None or not user.is_authenticated:
            logger.info('Authentication failed, logout.')
            auth.logout(request)
            return

        backend = auth.load_backend(user.backend)
        if not isinstance(backend, UniversalAuthBackend):
            logger.info("User is not validate by UniversalAuthBackend, skip login processes.")
            return

        logger.debug('Authentication finished, username: %s', user.username)
        request.session['provider_type'] = user.provider_type.value
        request.session['bkpaas_user_id'] = user.bkpaas_user_id
        request.session['bkpaas_authenticated_at'] = time.time()
        request.session['auth_credentials'] = credentials
        # python3 compatibility
        request.session['user_token'] = force_str(pickle.dumps(user.token), 'latin1')

        # Calling `auth.login` will rotate CSRF token and modify user session, only do this when the authenticated
        # user was different with the user stored in session. Otherwise CSRF token validation may fail due to the
        # rotation.
        if getattr(request, "user", None) != user:
            auth.login(request, user)



# -*- coding: utf-8 -*-
"""Exceptions for bkpaas_auth module
"""


class ServiceError(Exception):
    """Login or Token service is not available"""


class HttpRequestError(Exception):
    """http request error"""


class ResponseError(Exception):
    """service response error"""


class InvalidSkeyError(Exception):
    """Invalid uin/skey given"""


class InvalidTokenCredentialsError(Exception):
    """When invalid credentials are given when exchange access token"""


class AccessPermissionDenied(Exception):
    """Authenticated user has no access permissions"""



# -*- coding: utf-8 -*-
from enum import IntEnum


class ProviderType(IntEnum):
    UIN = 1
    RTX = 2
    BK = 3
    DATABASE = 9

    def get_id_prefix(self):
        return '{0:02d}'.format(self.value)


# 蓝鲸统一登录约定的错误码, 表示用户认证成功，但用户无应用访问权限
ACCESS_PERMISSION_DENIED_CODE = 1302403



# -*- coding: utf-8 -*-

class ARC4:
    """ARC4 算法的实现

    注意：ARC4 加密算法很容易被破解，目前项目中仅用于解编码（类似 base64 的作用）
    """

    def __init__(self, key):
        assert (isinstance(key, (bytes, bytearray)))

        s = list(range(0x100))
        j = 0
        for i in range(0x100):
            j = (s[i] + key[i % len(key)] + j) & 0xff
            s[i], s[j] = s[j], s[i]

        self.s = s
        self.key_stream = self._key_stream_generator()

    def encrypt(self, data):
        """加密数据"""
        return self._crypt(data)

    def decrypt(self, data):
        """解密数据"""
        return self._crypt(data)

    def _crypt(self, data):
        assert (isinstance(data, (bytes, bytearray)))
        return bytes([a ^ b for a, b in zip(data, self.key_stream)])

    def _key_stream_generator(self):
        s = self.s.copy()
        x = y = 0
        while True:
            x = (x + 1) & 0xff
            y = (s[x] + y) & 0xff
            s[x], s[y] = s[y], s[x]
            i = (s[x] + s[y]) & 0xff
            yield s[i]



# -*- coding: utf-8 -*-
"""
http基础方法

Rules:
1. POST/DELETE/PUT: json in
2. GET带参数，HEAD不带参数
3. 所有请求 json out，如果resp.json报错, 则是接口问题
"""
import json
import logging
from typing import Union

import requests

from bkpaas_auth.conf import bkauth_settings
from bkpaas_auth.core.exceptions import HttpRequestError, ServiceError
from bkpaas_auth.utils import scrub_data

logger = logging.getLogger(__name__)

request_adapter = requests.adapters.HTTPAdapter(pool_connections=20, pool_maxsize=20)


def get_requests_session():
    """Return an empty requests session, use the function to reuse HTTP connections"""
    session = requests.session()
    session.mount("http://", request_adapter)
    session.mount("https://", request_adapter)
    session.verify = bkauth_settings.REQUESTS_VERIFY
    session.cert = bkauth_settings.REQUESTS_CERT
    return session


def build_req_details_str(method, url, params, data, **kwargs) -> str:
    """Build the request details string for logging purpose."""
    msg = f'{method} {url}'
    if params:
        msg += f', params: {scrub_data(params)}'
    if data:
        msg += f', data: {scrub_data(data)}'
    msg += f', kwargs: {kwargs}'
    return msg


def _http_request(method: str, url: str, **kwargs) -> requests.Response:
    session = get_requests_session()
    params = kwargs.pop("params", None)
    data = kwargs.pop("data", None)

    req_details = build_req_details_str(method, url, params, data, **kwargs)
    logger.debug("Sending HTTP request, req details: %s", req_details)

    try:
        resp = session.request(method, url, params=params, data=data, **kwargs)
    except requests.exceptions.RequestException as e:
        logger.exception("http request error! req details: %s", req_details)
        raise HttpRequestError(f"http request error: {e}")

    return resp


def resp_to_json(resp: requests.Response) -> Union[dict, list]:
    try:
        return resp.json()
    except json.decoder.JSONDecodeError:
        logger.exception(
            "response json error! response.status_code: %s, response.text: %s",
            resp.status_code,
            resp.text,
        )
        raise ServiceError("parse json response error")


def http_get(url: str, **kwargs) -> requests.Response:
    return _http_request(method="GET", url=url, **kwargs)



# -*- coding: utf-8 -*-
"""Access token for blueking
"""
import datetime
import json
import logging
from abc import abstractmethod
from typing import NamedTuple

from django.utils.timezone import now
from django.utils.translation import get_language

from bkpaas_auth.conf import bkauth_settings
from bkpaas_auth.core.constants import ACCESS_PERMISSION_DENIED_CODE, ProviderType
from bkpaas_auth.core.exceptions import (
    AccessPermissionDenied,
    HttpRequestError,
    InvalidTokenCredentialsError,
    ResponseError,
    ServiceError,
)
from bkpaas_auth.core.http import http_get, resp_to_json
from bkpaas_auth.core.services import get_app_credentials
from bkpaas_auth.core.user_info import BkUserInfo, RtxUserInfo, UserInfo
from bkpaas_auth.models import User
from bkpaas_auth.utils import scrub_data

logger = logging.getLogger(__name__)


class UserAccount(NamedTuple):
    """
    用户账号信息

    :param bk_username: 用户唯一标识，全局唯一
    :param display_name: 用户展示名.
    :param tenant_id: 用户所属租户 ID. 为 None 时表示用户无租户信息(可能接口不支持)
    """

    bk_username: str
    display_name: str
    tenant_id: str | None = None


class AbstractRequestBackend:
    def request_user_account(self, **credentials) -> UserAccount:
        """Get user account through credentials

        :raises ServiceError: When the backend service is not available.
        :raises ResponseError: When the backend response status code is not in the 20x range, nor 403.
        :raises AccessPermissionDenied: When the user does not have access permission.
        :raises InvalidTokenCredentialsError: When invalid credentials are provided.
        """
        if bkauth_settings.USER_INFO_APIGW_URL:
            return self._request_apigw(**credentials)

        return self._request_esb(**credentials)

    @staticmethod
    @abstractmethod
    def _request_apigw(**credentials):
        """get user account by apigw"""

    @staticmethod
    @abstractmethod
    def _request_esb(**credentials):
        """get user account by esb"""


class TokenRequestBackend(AbstractRequestBackend):

    provider_type = ProviderType.BK

    @staticmethod
    def _request_apigw(**credentials) -> UserAccount:
        try:
            resp = http_get(
                bkauth_settings.USER_INFO_APIGW_URL,
                timeout=10,
                headers={
                    'blueking-language': get_language(),
                    "X-Bkapi-Authorization": json.dumps(get_app_credentials()),
                    # 全租户应用，调用全租户网关时，网关会强制要求传递 X-Bk-Tenant-Id, 但不会实际校验值的有效性, 统一传 default
                    "X-Bk-Tenant-Id": "default",
                },
                params=credentials,
            )
        except HttpRequestError:
            raise ServiceError('Unable to request services')

        resp_json = resp_to_json(resp)

        if not isinstance(resp_json, dict):
            raise ValueError(f'response type expect dict, got: {resp_json}')

        if resp.status_code == 200:
            bk_username = resp_json["data"]["bk_username"]
            return UserAccount(
                bk_username=bk_username,
                display_name=resp_json["data"].get("display_name") or bk_username,
                tenant_id=resp_json["data"].get("tenant_id"),
            )

        if resp.status_code == 403:
            raise AccessPermissionDenied(resp_json["error"]["message"])

        raise ResponseError(resp_json["error"]["message"])

    @staticmethod
    def _request_esb(**credentials) -> UserAccount:
        try:
            resp = http_get(
                bkauth_settings.USER_COOKIE_VERIFY_URL,
                timeout=10,
                headers={
                    'blueking-language': get_language(),
                    "X-Bkapi-Authorization": json.dumps(dict(credentials, **get_app_credentials())),
                },
                params=credentials,
            )
        except HttpRequestError:
            raise ServiceError('unable to fetch token services')

        resp_json = resp_to_json(resp)

        if not isinstance(resp_json, dict):
            raise ValueError(f'response type expect dict, got: {resp_json}')

        # API 返回格式为：{"result": true, "code": 0, "message": "", "data": {"bk_username": "xxx"}}
        code = resp_json.get('code')
        if code == 0:
            username = resp_json["data"]["bk_username"]
            return UserAccount(bk_username=username, display_name=username)

        logger.debug(
            f'Get user fail, url: {bkauth_settings.USER_COOKIE_VERIFY_URL}, '
            f'params: {scrub_data(credentials)}, response: {resp_json}'
        )

        # 用户认证成功，但用户无应用访问权限
        if code == ACCESS_PERMISSION_DENIED_CODE:
            raise AccessPermissionDenied(resp_json.get('message'))

        raise InvalidTokenCredentialsError('Invalid credentials given')


class RequestBackend(AbstractRequestBackend):

    provider_type = ProviderType.RTX

    @staticmethod
    def _request_esb(**credentials) -> UserAccount:
        try:
            resp = http_get(bkauth_settings.USER_COOKIE_VERIFY_URL, params=credentials, timeout=10)
        except HttpRequestError:
            raise ServiceError('unable to fetch token services')

        resp_json = resp_to_json(resp)

        if not isinstance(resp_json, dict):
            raise ValueError(f'response type expect dict, got: {resp_json}')

        # API 返回格式为：{"msg": "", "data": {"username": "xxx"}, "ret": 0}
        if resp_json.get('ret') != 0:
            logger.debug(
                f'Get user fail, url: {bkauth_settings.USER_COOKIE_VERIFY_URL}, '
                f'params: {scrub_data(credentials)}, response: {resp_json}'
            )
            raise InvalidTokenCredentialsError('Invalid credentials given')

        username = resp_json["data"]["username"]
        return UserAccount(bk_username=username, display_name=username)

    @staticmethod
    def _request_apigw(**credentials):
        raise NotImplementedError("No APIGW for RTX Backend")


class LoginToken:
    """Access token object"""

    token_timeout_margin = 300

    def __init__(self, login_token=None, expires_in=None):
        assert login_token, 'Must provide token string'
        assert expires_in, 'Must provide expires_in seconds'
        self.login_token = login_token
        self.expires_at = now() + datetime.timedelta(seconds=expires_in)
        self.issued_at = now()
        self.user_info = UserInfo(username='AnonymousUser')

    def __str__(self):
        return 'token: {} expires_at: {}'.format(self.login_token, self.expires_at)

    def expired(self):
        return self.expires_at < now()

    def make_user(self, provider_type):
        self.user_info.provider_type = provider_type
        return create_user_from_token(self)


def mocked_create_user_from_token(
    token: LoginToken, provider_type: int = ProviderType.RTX, username: str = bkauth_settings.MOCKED_USER_NAME
) -> User:
    """Mocked create_user function, only for temporary use"""
    if provider_type == ProviderType.RTX:
        token.user_info = RtxUserInfo(
            LoginName=username,
            ChineseName=username,
        )
    elif provider_type == ProviderType.BK:
        token.user_info = BkUserInfo(bk_username=username, chname=username, email='', phone='')
    else:
        raise ValueError('Invalid provider_type given.')
    return create_user_from_token(token)


def create_user_from_token(token: LoginToken) -> User:
    """Create a user object from user info dict"""
    user = User(token=token)
    return token.user_info.provide(user)



# -*- coding: utf-8 -*-
import binascii
from typing import Tuple, Union

from six import ensure_binary, ensure_text

from bkpaas_auth.core.constants import ProviderType
from bkpaas_auth.core.algorithms import ARC4


class BluekingUserIdEncoder:
    """Generator for blueking user id"""

    # It is not a real secret key, just for encoding the username
    secret_key = 'jdvoqu3o4'

    def encode(self, provider_type: Union[int, ProviderType], username: Union[str, bytes]):
        """Generate a hex string used as blueking user id

        :param provider_type: See constants.ProviderType
        :param username: User uin or user rtx username
        :returns: A hex string
        """
        id_prefix = ProviderType(provider_type).get_id_prefix()

        encoded = ARC4(ensure_binary(self.secret_key)).encrypt(ensure_binary(username))
        return id_prefix + ensure_text(binascii.hexlify(encoded))

    def decode(self, user_id: Union[str, bytes]) -> Tuple[int, str]:
        """Decode a given bk_user_id to the combination of (provider_type, username)

        :param str user_id: Blueking user id
        :returns: (provider_type, username)
        """
        _provider_type, username = user_id[:2], user_id[2:]
        provider_type = int(_provider_type)

        decoded = ARC4(ensure_binary(self.secret_key)).decrypt(binascii.unhexlify(username))
        return provider_type, ensure_text(decoded)


user_id_encoder = BluekingUserIdEncoder()
